-----------------------------------------------------
----Show us Three Dimensions--STD (Render SubLib)----
-----------------------------------------------------

vectors = require "STD.vector"

local STD = {}

--Utilize Vectors for camera manipulation
STD.camera = vectors.newVector(0,20,-20)
STD.camera.dir = vectors.newVector(0,0,0)
STD.camera.fov = 256



STD.objects = {
}

STD.showBorder = true

function STD.newObject()
	local o = {
		faces = {
		},
		newFace = newFace
	}
	table.insert(STD.objects,o)
	return o
end

function newFace(...)
	local self = arg[1]
	local f = {}
	if type(arg[2])=="number" then
		for i=2,(#arg),3 do
			if type(arg[i])=="number" then
			table.insert(f,vectors.newVector(arg[i],arg[i+1],arg[i+2]))
			end
		end
	elseif type(arg[2])=="table" then
		for i=2,#arg-1 do
			if arg[i].x~=nil then
				table.insert(f,vectors.newVector(arg[i].x,arg[i].y,arg[i].z))
			end
		end
	end
	if type(arg[#arg])=="table" then
		if arg[#arg].x==nil then
			f.col = arg[#arg]
		end
	end
	f.col = f.col or {math.random(255),math.random(255),math.random(255)}
	table.insert(self.faces,f)
end

local function zdex(facea,faceb)
	local avxa = 0
	local avya = 0
	local avza = 0
	local ta = 0
	local avxb = 0
	local avyb = 0
	local avzb = 0
	local tb = 0
	local faceb = faceb or facea
	
	for i=1,#facea do
		avxa = avxa+(facea[i].x)
		avya = avya+(facea[i].y)
		avza = avza+(facea[i].z)
		ta = ta+1
	end
	for i=1,#faceb do
		avxb = avxb+(faceb[i].x)
		avyb = avyb+(faceb[i].y)
		avzb = avzb+(faceb[i].z)
		tb = tb+1
	end
	avxa = avxa/ta
	avxb = avxb/tb
	avya = avya/ta
	avyb = avyb/tb
	avza = avza/ta
	avzb = avzb/tb
	
	local dista = math.sqrt((avxa-STD.camera.x)^2+(avya-STD.camera.y)^2+(avza-STD.camera.z)^2)
	local distb = math.sqrt((avxb-STD.camera.x)^2+(avyb-STD.camera.y)^2+(avzb-STD.camera.z)^2)

	return dista>distb
end

function STD.renderAll()
	local globalObj = {faces = {}}
	for k,v in pairs(STD.objects) do
		if type(k)=="number" then
			for i=1,#v.faces do
				table.insert(globalObj.faces,v.faces[i])
			end
		end
	end
	STD.renderObject(globalObj)
end

function STD.renderList(...)
	local obj = {faces = {}}
	for k,v in pairs(arg) do
		if type(k)=="number" then
			for i=1,#v.faces do
				table.insert(obj.faces,v.faces[i])
			end
		end
	end
	STD.renderObject(obj)
end

function STD.renderObject(obj)
	if STD.camera.dir.x>math.pi*2 then STD.camera.dir.x = STD.camera.dir.x-math.pi*2 end
	if STD.camera.dir.x<-math.pi*2 then STD.camera.dir.x = STD.camera.dir.x+math.pi*2 end
	if STD.camera.dir.y>math.pi*2 then STD.camera.dir.y = STD.camera.dir.y-math.pi*2 end
	if STD.camera.dir.y<-math.pi*2 then STD.camera.dir.y = STD.camera.dir.y+math.pi*2 end
	if STD.camera.dir.z>math.pi*2 then STD.camera.dir.z = STD.camera.dir.z-math.pi*2 end
	if STD.camera.dir.z<-math.pi*2 then STD.camera.dir.z = STD.camera.dir.z+math.pi*2 end
	table.sort(obj.faces,zdex)
	for f=1,#obj.faces do
		local points = {}
		local allRender = false
		love.graphics.setColor(255,255,255)
		for p=1,#obj.faces[f] do
			local np = #points
			local render = false
			points[np+1],points[np+2],render = STD.pointToScreen(obj.faces[f][p])
			allRender = allRender or render
		end
		if allRender then
			love.graphics.setColor(obj.faces[f].col)
			pcall(function() love.graphics.polygon("fill",points); if STD.showBorder then love.graphics.setColor(0,0,0); love.graphics.polygon("line",points) end end)
		end
	end
end

function STD.pointToScreen(p)
	local tp = {x = p.x, y = p.y, z = p.z}
	STD.rotateY(tp,STD.camera,STD.camera.dir.y)
	STD.rotateZ(tp,STD.camera,STD.camera.dir.z)
	STD.rotateX(tp,STD.camera,STD.camera.dir.x)
	local nx,ny,nz = tp.x-STD.camera.x,tp.y-STD.camera.y,tp.z-STD.camera.z
	
	local rz = nz
	nz = math.max(0.01,nz) --Hacky way to keep objects with missing edges (off screen) rendering properly
	
	local x,y = love.graphics.getWidth()/2 + nx * STD.camera.fov/nz, love.graphics.getWidth()/2 + ny *-1* STD.camera.fov/nz
	
	if rz<0 then

	end
	
	return x,y,rz>0
end


function STD.rotateX(p,c,r)
	local z = p.z - c.z 			
	local y = p.y - c.y
	local d = math.sqrt(y^2+z^2)
	local t = math.atan2(y,z) + r
	p.z = c.z + d * math.cos(t)
	p.y = c.y + d * math.sin(t)
end

function STD.rotateY(p,c,r)
	local z = p.z - c.z
	local x = p.x - c.x
	local d = math.sqrt(x^2+z^2)
	local t = math.atan2(x,z) + r
	p.z = c.z + d * math.cos(t)
	p.x = c.x + d * math.sin(t)
end

function STD.rotateZ(p,c,r)
	local x = p.x - c.x
	local y = p.y - c.y
	local d = math.sqrt(y^2+x^2)
	local t = math.atan2(y,x) + r
	p.x = c.x + d * math.cos(t)
	p.y = c.y + d * math.sin(t)
end

return STD